@@ -26,6 +26,9 @@ from video_diffusion_pytorch.text import tokenize, bert_embed, BERT_MODEL_DIM
 def exists(x):
     return x is not None
 
+def noop(*args, **kwargs):
+    pass
+
 def is_odd(n):
     return (n % 2) == 1
 
@@ -205,40 +208,6 @@ class ResnetBlock(nn.Module):
         h = self.block2(h)
         return h + self.res_conv(x)
 
-class ConvNextBlock(nn.Module):
-    """ https://arxiv.org/abs/2201.03545 """
-
-    def __init__(self, dim, dim_out, *, time_emb_dim = None, mult = 2):
-        super().__init__()
-        hidden_dim = int(dim_out * mult)
-
-        self.mlp = nn.Sequential(
-            nn.GELU(),
-            nn.Linear(time_emb_dim, dim)
-        ) if exists(time_emb_dim) else None
-
-        self.ds_conv = nn.Conv3d(dim, dim, (1, 7, 7), padding = (0, 3, 3), groups = dim)
-
-        self.net = nn.Sequential(
-            LayerNorm(dim),
-            nn.Conv3d(dim, hidden_dim, (1, 3, 3), padding = (0, 1, 1)),
-            nn.GELU(),
-            nn.Conv3d(hidden_dim, dim_out, (1, 3, 3), padding = (0, 1, 1))
-        )
-
-        self.res_conv = nn.Conv3d(dim, dim_out, 1) if dim != dim_out else nn.Identity()
-
-    def forward(self, x, time_emb = None):
-        h = self.ds_conv(x)
-
-        if exists(self.mlp):
-            assert exists(time_emb), 'time emb must be passed in'
-            condition = self.mlp(time_emb)
-            h = h + rearrange(condition, 'b c -> b c 1 1 1')
-
-        h = self.net(h)
-        return h + self.res_conv(x)
-
 class SpatialLinearAttention(nn.Module):
     def __init__(self, dim, heads = 4, dim_head = 32):
         super().__init__()
@@ -304,35 +273,62 @@ class Attention(nn.Module):
         self,
         x,
         pos_bias = None,
-        attend_self_only = False
+        focus_present_mask = None
     ):
+        n, device = x.shape[-2], x.device
+
         qkv = self.to_qkv(x).chunk(3, dim = -1)
 
-        if not attend_self_only:
-            q, k, v = rearrange_many(qkv, '... n (h d) -> ... h n d', h = self.heads)
-            q = q * self.scale
+        if exists(focus_present_mask) and focus_present_mask.all():
+            # if all batch samples are focusing on present
+            # it would be equivalent to passing that token's values through to the output
+            values = qkv[-1]
+            return self.to_out(values)
 
-            if exists(self.rotary_emb):
-                q = self.rotary_emb.rotate_queries_or_keys(q)
-                k = self.rotary_emb.rotate_queries_or_keys(k)
+        # split out heads
 
-            sim = einsum('... h i d, ... h j d -> ... h i j', q, k)
+        q, k, v = rearrange_many(qkv, '... n (h d) -> ... h n d', h = self.heads)
 
-            if exists(pos_bias):
-                sim = sim + pos_bias
+        # scale
 
-            sim = sim - sim.amax(dim = -1, keepdim = True).detach()
-            attn = sim.softmax(dim = -1)
+        q = q * self.scale
 
-            out = einsum('... h i j, ... h j d -> ... h i d', attn, v)
-            out = rearrange(out, '... h n d -> ... n (h d)')
-        else:
-            # it was not clear in the paper
-            # but i assume by arresting the attention across time
-            # would mean simply having each token attend to itself
-            # which would be equivalent to passing that token's values through to the output
-            out = qkv[-1]
+        # rotate positions into queries and keys for time attention
 
+        if exists(self.rotary_emb):
+            q = self.rotary_emb.rotate_queries_or_keys(q)
+            k = self.rotary_emb.rotate_queries_or_keys(k)
+
+        # similarity
+
+        sim = einsum('... h i d, ... h j d -> ... h i j', q, k)
+
+        # relative positional bias
+
+        if exists(pos_bias):
+            sim = sim + pos_bias
+
+        if exists(focus_present_mask) and not (~focus_present_mask).all():
+            attend_all_mask = torch.ones((n, n), device = device, dtype = torch.bool)
+            attend_self_mask = torch.eye(n, device = device, dtype = torch.bool)
+
+            mask = torch.where(
+                rearrange(focus_present_mask, 'b -> b 1 1 1 1'),
+                rearrange(attend_self_mask, 'i j -> 1 1 1 i j'),
+                rearrange(attend_all_mask, 'i j -> 1 1 1 i j'),
+            )
+
+            sim = sim.masked_fill(mask, -torch.finfo(sim.dtype).max)
+
+        # numerical stability
+
+        sim = sim - sim.amax(dim = -1, keepdim = True).detach()
+        attn = sim.softmax(dim = -1)
+
+        # aggregate values
+
+        out = einsum('... h i j, ... h j d -> ... h i d', attn, v)
+        out = rearrange(out, '... h n d -> ... n (h d)')
         return self.to_out(out)
 
 # model
@@ -352,8 +348,7 @@ class Unet3D(nn.Module):
         init_kernel_size = 7,
         use_sparse_linear_attn = True,
         block_type = 'resnet',
-        resnet_groups = 8,
-        convnext_mult = 2
+        resnet_groups = 8
     ):
         super().__init__()
         self.channels = channels
@@ -407,13 +402,7 @@ class Unet3D(nn.Module):
 
         # block type
 
-        if block_type == 'resnet':
-            block_klass = partial(ResnetBlock, groups = resnet_groups)
-        elif block_type == 'convnext':
-            block_klass = partial(ConvNextBlock, mult = convnext_mult)
-        else:
-            raise ValueError(f'unknown block type {block_type}')
-
+        block_klass = partial(ResnetBlock, groups = resnet_groups)
         block_klass_cond = partial(block_klass, time_emb_dim = cond_dim)
 
         # modules for all layers
@@ -475,9 +464,12 @@ class Unet3D(nn.Module):
         time,
         cond = None,
         null_cond_prob = 0.,
-        focus_on_the_present = True  # sounds quite spiritual
+        prob_focus_present = 0.  # probability at which a given batch sample will focus on the present (0. is all off, 1. is completely arrested attention across time)
     ):
         assert not (self.has_cond and not exists(cond)), 'cond must be passed in if cond_dim specified'
+        batch, device = x.shape[0], x.device
+
+        focus_present_mask = prob_mask_like((batch,), prob_focus_present, device = device)
 
         x = self.init_conv(x)
 
@@ -499,13 +491,13 @@ class Unet3D(nn.Module):
             x = block1(x, t)
             x = block2(x, t)
             x = spatial_attn(x)
-            x = temporal_attn(x, pos_bias = time_rel_pos_bias, attend_self_only = focus_on_the_present)
+            x = temporal_attn(x, pos_bias = time_rel_pos_bias, focus_present_mask = focus_present_mask)
             h.append(x)
             x = downsample(x)
 
         x = self.mid_block1(x, t)
         x = self.mid_spatial_attn(x)
-        x = self.mid_temporal_attn(x, pos_bias = time_rel_pos_bias, attend_self_only = focus_on_the_present)
+        x = self.mid_temporal_attn(x, pos_bias = time_rel_pos_bias, focus_present_mask = focus_present_mask)
         x = self.mid_block2(x, t)
 
         for block1, block2, spatial_attn, temporal_attn, upsample in self.ups:
@@ -513,7 +505,7 @@ class Unet3D(nn.Module):
             x = block1(x, t)
             x = block2(x, t)
             x = spatial_attn(x)
-            x = temporal_attn(x, pos_bias = time_rel_pos_bias, attend_self_only = focus_on_the_present)
+            x = temporal_attn(x, pos_bias = time_rel_pos_bias, focus_present_mask = focus_present_mask)
             x = upsample(x)
 
         return self.final_conv(x)
@@ -626,7 +618,7 @@ class GaussianDiffusion(nn.Module):
         model_mean, posterior_variance, posterior_log_variance = self.q_posterior(x_start=x_recon, x_t=x, t=t)
         return model_mean, posterior_variance, posterior_log_variance
 
-    @torch.no_grad()
+    @torch.inference_mode()
     def p_sample(self, x, t, cond = None, cond_scale = 1., clip_denoised = True, repeat_noise = False):
         b, *_, device = *x.shape, x.device
         model_mean, _, model_log_variance = self.p_mean_variance(x = x, t = t, clip_denoised = clip_denoised, cond = cond, cond_scale = cond_scale)
@@ -635,7 +627,7 @@ class GaussianDiffusion(nn.Module):
         nonzero_mask = (1 - (t == 0).float()).reshape(b, *((1,) * (len(x.shape) - 1)))
         return model_mean + nonzero_mask * (0.5 * model_log_variance).exp() * noise
 
-    @torch.no_grad()
+    @torch.inference_mode()
     def p_sample_loop(self, shape, cond = None, cond_scale = 1.):
         device = self.betas.device
 
@@ -646,7 +638,7 @@ class GaussianDiffusion(nn.Module):
             img = self.p_sample(img, torch.full((b,), i, device=device, dtype=torch.long), cond = cond, cond_scale = cond_scale)
         return img
 
-    @torch.no_grad()
+    @torch.inference_mode()
     def sample(self, cond = None, cond_scale = 1., batch_size = 16):
         device = next(self.denoise_fn.parameters()).device
 
@@ -659,7 +651,7 @@ class GaussianDiffusion(nn.Module):
         num_frames = self.num_frames
         return self.p_sample_loop((batch_size, channels, num_frames, image_size, image_size), cond = cond, cond_scale = cond_scale)
 
-    @torch.no_grad()
+    @torch.inference_mode()
     def interpolate(self, x1, x2, t = None, lam = 0.5):
         b, *_, device = *x1.shape, x1.device
         t = default(t, self.num_timesteps - 1)
@@ -803,7 +795,6 @@ class Dataset(data.Dataset):
 
 # trainer class
 
-import wandb
 class Trainer(object):
     def __init__(
         self,
@@ -895,8 +886,11 @@ class Trainer(object):
 
     def train(
         self,
-        focus_on_the_present = False
+        prob_focus_present = 0.,
+        log_fn = noop
     ):
+        assert callable(log_fn)
+
         while self.step < self.train_num_steps:
             for i in range(self.gradient_accumulate_every):
                 data = next(self.dl).cuda()
@@ -904,7 +898,7 @@ class Trainer(object):
                 with autocast(enabled = self.amp):
                     loss = self.model(
                         data,
-                        focus_on_the_present = focus_on_the_present
+                        prob_focus_present = prob_focus_present
                     )
 
                     self.scaler.scale(loss / self.gradient_accumulate_every).backward()
@@ -938,10 +932,10 @@ class Trainer(object):
                 one_gif = rearrange(all_videos_list, '(i j) c f h w -> c f (i h) (j w)', i = self.num_sample_rows)
                 video_path = str(self.results_folder / str(f'{milestone}.gif'))
                 video_tensor_to_gif(one_gif, video_path)
-                log = {**log, 'sample': wandb.Video(video_path)}
+                log = {**log, 'sample': video_path}
                 self.save(milestone)
 
-            wandb.log(log)
+            log_fn(log)
             self.step += 1
 
         print('training completed')
